% Compute EEG leadfield using the standard (CG-) FEM approach,
% in a 4-layer spherical tetrahedral sphere model
% with different source models (Partial integration, St. Venant, Whitney and Subtraction)
% using the transfer matrix approach

% I. add path to duneuro-matlab
addpath(genpath('/path/to/DUNEuro-matlab'));

% II. Define input files
path_input              = '/path/to/input/';
path_ouput              = '/path/to/output/';
filename_grid           = fullfile(path_input, 'sphere_tet_mesh_4c.msh');
filename_tensors        = fullfile(path_input, 'sphere_4c.cond');
filename_electrodes     = fullfile(path_input, 'sphere_electrodes.txt');
filename_dipoles        = fullfile(path_input, 'sphere_dipoles.txt');
filename_analytical     = fullfile(path_input, 'sphere_eeg_analytical.txt');

% III. Create MEEG driver
% We create the driver object which will read the mesh along with the conductivity tensors from the provided files
cfg = [];
cfg.solver_type = 'cg';
cfg.type = 'fitted';
cfg.element_type = 'tetrahedron';
cfg.volume_conductor.grid.filename = filename_grid;
cfg.volume_conductor.tensors.filename = filename_tensors;
driver = duneuro_meeg(cfg);

% IV. Read and set electrode positions
% When projecting the electrodes, we choose the closest nodes
electrodes = table2array(readtable(filename_electrodes))';
cfg = [];
cfg.codims = '3';
cfg.type = 'closest_subentity_center';
driver.set_electrodes(electrodes, cfg);

% V. Compute EEG leadfield
% Compute the transfer matrix
cfg = [];
cfg.solver.reduction = '1e-10';
transfer_matrix = driver.compute_eeg_transfer_matrix(cfg);

% Create source model configurations (Partial integration St. Venant, Subtraction, Whitney)
cfg_pi = struct('type','partial_integration');
cfg_venant =  struct('type','venant',...
    'numberOfMoments', '3', ...
    'referenceLength' , '20', ...
    'weightingExponent' , '1', ...
    'relaxationFactor' , '1e-6', ...
    'mixedMoments' , 'True', ...
    'restrict' , 'True', ...
    'initialization' , 'closest_vertex');

cfg_whitney =  struct('type','whitney',...
    'referenceLength' , '20', ...
    'restricted' , 'True', ...
    'faceSources' , 'all', ...
    'edgeSources'  , 'all', ...
    'interpolation' , 'PBO');

cfg_subtraction = struct('type','subtraction', ...
    'intorderadd', '2', ...
    'intorderadd_lb', '2');

configs_source = {cfg_pi, cfg_venant, cfg_whitney, cfg_subtraction};

% Read dipoles
dipoles                 = table2array(readtable(filename_dipoles))';

% Apply the transfer matrix
for i=1:size(configs_source,2)
    cfg = [];
    cfg.post_process = 'True';
    cfg.subtract_mean = 'True';
    cfg.source_model = configs_source{i};
    lf{i}  = driver.apply_eeg_transfer(transfer_matrix, dipoles, cfg);
end

% print list of relevant publications
driver.print_citations();
