# Compute EEG leadfield using the Discontinuous Galerkin (DG-) FEM approach,
# in a 4-layer spherical tetrahedral sphere model
# with the partial integration source model
# using the transfer matrix apporach

# I. Import libraries
import numpy as np
import pandas as pd
import seaborn as sb
import matplotlib.pyplot as plt
import duneuropy as dp
import os

# II. Define input files
folder_input = '/path/to/input/'
folder_output = '/path/to/output/'
filename_electrodes = os.path.join(folder_input, 'sphere_electrodes.txt')
filename_grid = os.path.join(folder_input, 'sphere_tet_mesh_4c.msh')
filename_tensors = os.path.join(folder_input, 'sphere_4c.cond')
filename_dipoles = os.path.join(folder_input, 'sphere_dipoles.txt')
filename_analytical = os.path.join(folder_input, 'sphere_eeg_analytical.txt')

# III. Create MEEG driver
# We create the driver object which will read the mesh along with the conductivity tensors from the provided files
config = {
    'type' : 'fitted',
    'solver_type' : 'dg',
    'element_type' : 'tetrahedron',
    'volume_conductor' : {
        'grid.filename' : filename_grid,
        'tensors.filename' : filename_tensors
    },
    'solver' : {
        'edge_norm_type' : 'houston',
        'penalty' : 20,
        'reduction' : 1e-10,
        'scheme' : 'sipg',
        'weights' : 'tensorOnly',

    }
}
driver = dp.MEEGDriver3d(config)

# IV. Read and set electrode positions
# When projecting the electrodes, we choose the closest nodes
electrodes = np.genfromtxt(filename_electrodes,delimiter=None) 
electrodes = [dp.FieldVector3D(t) for t in electrodes.tolist()]
electrode_config = {
    'type' : 'closest_subentity_center',
    'codims' : [3]
}
driver.setElectrodes(electrodes, electrode_config)

# V. Compute EEG leadfield
# Compute the transfer matrix
configTM = {
    'solver' : {
        'reduction' : 1e-10
    }
}
tm = driver.computeEEGTransferMatrix(configTM)
tm_eeg = np.array(tm[0])

# Create source model configurations (partial integration source model)
source_model_configs = {
    'Partial integration' : {
        'type' : 'partial_integration'
    }
}

# Read dipoles
dipoles = dp.read_dipoles_3d(filename_dipoles)

# Apply the transfer matrix
solutions = dict()
for sm in source_model_configs:
    lf = driver.applyEEGTransfer(tm_eeg, dipoles, {
                    'source_model' : source_model_configs[sm],
                    'post_process' : True,
                    'subtract_mean' : True
                })
    solutions[sm] = np.array(lf[0])

# VI. Comparison with (quasi-)analytical solution
# load quasi-analytical solution
analytical = np.transpose(np.genfromtxt(filename_analytical,delimiter= ''))

# compute errors between numerical and analytical solution:
# RDM(u,v) = \left\|\frac{u}{\|u\|}-\frac{v}{\|v\|}\right\|
# MAG(u,v) = \frac{\|u\|}{\|v\|}
# lnMAG(u,v)= \ln(MAG(u,v))
norm = lambda x : np.linalg.norm(x, axis=1)
nnum = {sm : norm(solutions[sm]) for sm in solutions}
nana = norm(analytical)
lnmag = {sm: np.log(nnum[sm]/nana)  for sm in nnum}
rdm = {sm: norm(solutions[sm]/nnum[sm][:, None]-analytical/nana[:, None]) for sm in solutions}

# Plot errors
eccentricities = [np.transpose(np.round(np.linalg.norm(np.array(dip.position())-(127,127,127))/78,3)) for dip in dipoles]

df = pd.concat([pd.DataFrame({
        'eccentricity' : eccentricities,
        'RDM' : rdm[sm],
        'lnMAG' : lnmag[sm],
        'Source model' : sm,

    }) for sm in rdm])

plt.figure(figsize=(10,8))
sb.boxplot(y='RDM', x='eccentricity', 
                 data=df, 
                 palette="colorblind",
                 hue='Source model')
plt.savefig(os.path.join(folder_output, 'sphere_dg_tet_transfer_rdm.png'))

plt.figure(figsize=(10,8))
sb.boxplot(y='lnMAG', x='eccentricity', 
                 data=df, 
                 palette="colorblind",
                 hue='Source model')
plt.savefig(os.path.join(folder_output, 'sphere_dg_tet_transfer_mag.png'))

# Print a list of relevant publications
driver.print_citations()
