# Compute EEG leadfield using the Unfitted Discontinuous Galerkin (UDG-) FEM approach,
# in a 4-layer spherical sphere model
# with the partial integration source model
# using the transfer matrix apporach

# I. Import libraries
import numpy as np
import pandas as pd
import seaborn as sb
import matplotlib.pyplot as plt
import duneuropy as dp
import os

# II. Define input files
folder_input = '/path/to/input/'
folder_output = '/path/to/output/'
filename_electrodes = os.path.join(folder_input, 'sphere_electrodes.txt')
filename_dipoles = os.path.join(folder_input, 'sphere_dipoles.txt')
filename_analytical = os.path.join(folder_input, 'sphere_eeg_analytical.txt')

# III. Create MEEG driver
# We create the driver object using the sphere model characteristics and conductivities.
config = {
    'type' : 'unfitted',
    'solver_type' : 'udg',
    'compartments': 4,
    'volume_conductor' : {
        'grid' : {
            'cells' : (30, 30, 30),
            'upper_right' : (220, 220, 220),
            'lower_left' : (30, 30, 30),
            'refinements' : 0
        }
    },
    'domain' : {
        'domains' : ['skin', 'skull', 'csf', 'brain'],
        'level_sets' : ['outer_skin', 'skin_skull', 'skull_csf', 'csf_brain'],
        'brain.positions' :  ['iiii'],
        'csf_brain.radius' : 78,
        'csf.positions' : 'iiie',
        'skull_csf.radius' : 80,
        'skull.positions' : 'iiee',
        'skin_skull.radius' : 86,
        'skin.positions' : 'ieee',
        'outer_skin.radius' : 92
    },
    'solver' : {
        'conductivities' : [0.00043, 0.00001, 0.00179, 0.00033],
        'edge_norm_type' : 'face',
        'penalty' : 16,
        'intorderadd' : 0,
        'reduction' : 1e-10,
        'scheme' : 'sipg',
        'weights' : 'tensorOnly',
        'smoother' : 'default'
    }
}

for k in config['domain']['level_sets']:
    config['domain'][k] = {
        'type' : 'sphere',
        'center' : (127, 127, 127)
    }
driver = dp.MEEGDriver3d(config)

# IV. Read and set electrode positions
electrodes = np.genfromtxt(filename_electrodes,delimiter=None) 
electrodes = [dp.FieldVector3D(t) for t in electrodes.tolist()]
driver.setElectrodes(electrodes, {})

# V. Compute EEG leadfield
# Compute the transfer matrix
configTM = {
    'solver' : {
        'reduction' : 1e-10,
        'compartment' : 0
    }
}
tm = driver.computeEEGTransferMatrix(configTM)
tm_eeg = np.array(tm[0])

# Create source model configurations (partial integration source model)
source_model_configs = {
    'Partial integration' : {
        'type' : 'partial_integration',
        'compartment' : '3'
    },
}

# Read dipoles
dipoles = dp.read_dipoles_3d(filename_dipoles)

# Apply the transfer matrix
solutions = dict()
for sm in source_model_configs:
    lf = driver.applyEEGTransfer(tm_eeg, dipoles, {
                    'source_model' : source_model_configs[sm],
                    'post_process' : True,
                    'subtract_mean' : True
                })
    solutions[sm] = np.array(lf[0])

# VI. Comparison with (quasi-)analytical solution
# Load the analytical solution
analytical = np.transpose(np.genfromtxt(filename_analytical,delimiter= ''))

# Compute errors between numerical and analytical solution:
# RDM(u,v) = \left\|\frac{u}{\|u\|}-\frac{v}{\|v\|}\right\|
# MAG(u,v) = \frac{\|u\|}{\|v\|}
# lnMAG(u,v)= \ln(MAG(u,v))
norm = lambda x : np.linalg.norm(x, axis=1)
nnum = {sm : norm(solutions[sm]) for sm in solutions}
nana = norm(analytical)
lnmag = {sm: np.log(nnum[sm]/nana)  for sm in nnum}
rdm = {sm: norm(solutions[sm]/nnum[sm][:, None]-analytical/nana[:, None]) for sm in solutions}

# Plot errors
eccentricities = [np.transpose(np.round(np.linalg.norm(np.array(dip.position())-(127,127,127))/78,3)) for dip in dipoles]

df = pd.concat([pd.DataFrame({
        'eccentricity' : eccentricities,
        'RDM' : rdm[sm],
        'lnMAG' : lnmag[sm],
        'Source model' : sm,

    }) for sm in rdm])

plt.figure(figsize=(10,8))
sb.boxplot(y='RDM', x='eccentricity', 
                 data=df, 
                 palette="colorblind",
                 hue='Source model')
plt.savefig(os.path.join(folder_output, 'sphere_udg_transfer_rdm.png'))


plt.figure(figsize=(10,8))
sb.boxplot(y='lnMAG', x='eccentricity', 
                 data=df, 
                 palette="colorblind",
                 hue='Source model')
plt.savefig(os.path.join(folder_output, 'sphere_udg_transfer_mag.png'))

#Visualization of tissue surfaces, the first dipole and the resulting potential
driver.write({
    'format' : 'vtk',
    'filename' : os.path.join(folder_output,'sphere_udg_transfer_headmodel'),
    'mode' : 'boundary'
})
pvtk = dp.PointVTKWriter3d(dipoles[0])
pvtk.write(os.path.join(folder_output, 'sphere_udg_transfer_testdipole'))

pvtk = dp.PointVTKWriter3d(electrodes, True)
pvtk.addScalarData('potential', solutions['Partial integration'][0])
pvtk.write(os.path.join(folder_output,'sphere_udg_transfer_lf_pi'))

# Print a list of relevant publications
driver.print_citations()
